/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Collections.Generic;
using System.Drawing;

namespace ImageProc
{
	/// <summary>
	/// This class shall keep all the functionality for capturing 
	/// the desktop.
	/// </summary>
	public class CaptureWindow
	{
        public enum WINDOWTYPE {PW_WINDOW, PW_CLIENT};

		#region Public Class Functions
		public static Bitmap GetScreenImage()
		{
			//In size variable we shall keep the size of the screen.
            int nWidth, nHeight;
			
			//Variable to keep the handle to bitmap.
			IntPtr hBitmap;

			//Here we get the handle to the desktop device context.
            IntPtr hDC = GDI32.CreateDC("DISPLAY", null, null, IntPtr.Zero);

			//Here we make a compatible device context in memory for screen device context.
			IntPtr hMemDC = GDI32.CreateCompatibleDC(hDC);
			
			//We pass SM_CXSCREEN constant to GetSystemMetrics to get the X coordinates of screen.
			nWidth = USER32.GetSystemMetrics(USER32.SM_CXSCREEN);

			//We pass SM_CYSCREEN constant to GetSystemMetrics to get the Y coordinates of screen.
			nHeight = USER32.GetSystemMetrics(USER32.SM_CYSCREEN);
			
			//We create a compatible bitmap of screen size using screen device context.
			hBitmap = GDI32.CreateCompatibleBitmap(hDC, nWidth, nHeight);

			//As hBitmap is IntPtr we can not check it against null. For this purspose IntPtr.Zero is used.
			if (hBitmap!=IntPtr.Zero)
			{
				//Here we select the compatible bitmap in memeory device context and keeps the refrence to Old bitmap.
				IntPtr hOld = (IntPtr) GDI32.SelectObject(hMemDC, hBitmap);
				//We copy the Bitmap to the memory device context.
				GDI32.BitBlt(hMemDC, 0, 0, nWidth, nHeight, hDC, 0, 0, GDI32.SRCCOPY);
				//We select the old bitmap back to the memory device context.
				GDI32.SelectObject(hMemDC, hOld);
				//We delete the memory device context.
				GDI32.DeleteDC(hMemDC);
				//We release the screen device context.
                GDI32.DeleteDC(hDC);
				//Image is created by Image bitmap handle and stored in local variable.
				Bitmap bmp = System.Drawing.Image.FromHbitmap(hBitmap); 
				//Release the memory for compatible bitmap.
				GDI32.DeleteObject(hBitmap);
				//This statement runs the garbage collector manually.
				GC.Collect();
				//Return the bitmap 
				return bmp;
			}
		
			//If hBitmap is null retunrn null.
			return null;
		}

        public static Bitmap GetScreenImage(Rectangle rect)
        {
            // Coordinates of the rectangle to grab, and the size of the screen. 
            int nX1, nY1, nX2, nY2, nWidth, nHeight;

            // Get the coordinates
            nX1 = rect.Left;
            nY1 = rect.Top;
            nX2 = rect.Right;
            nY2 = rect.Bottom;

            //Variable to keep the handle to bitmap.
            IntPtr hBitmap;

            //Here we get the handle to the desktop device context.
            IntPtr hDC = USER32.GetDC(USER32.GetDesktopWindow());

            //Here we make a compatible device context in memory for screen device context.
            IntPtr hMemDC = GDI32.CreateCompatibleDC(hDC);

            //We pass SM_CXSCREEN constant to GetSystemMetrics to get the X coordinates of screen.
            nWidth = USER32.GetSystemMetrics(USER32.SM_CXSCREEN);

            //We pass SM_CYSCREEN constant to GetSystemMetrics to get the Y coordinates of screen.
            nHeight = USER32.GetSystemMetrics(USER32.SM_CYSCREEN);

            // Get the valid screen area
            if (nX1 < 0) nX1 = 0;
            if (nY1 < 0) nY1 = 0;
            if (nX2 > nWidth) nX2 = nWidth;
            if (nY2 > nHeight) nY2 = nHeight;

            nWidth = nX2 - nX1;
            nHeight = nY2 - nY1; 

            //We create a compatible bitmap of screen size using screen device context.
            hBitmap = GDI32.CreateCompatibleBitmap(hDC, nWidth, nHeight);

            //As hBitmap is IntPtr we can not check it against null. For this purspose IntPtr.Zero is used.
            if (hBitmap != IntPtr.Zero)
            {
                //Here we select the compatible bitmap in memeory device context and keeps the refrence to Old bitmap.
                IntPtr hOld = (IntPtr)GDI32.SelectObject(hMemDC, hBitmap);
                //We copy the Bitmap to the memory device context.
                GDI32.BitBlt(hMemDC, 0, 0, nWidth, nHeight, hDC, nX1, nY1, GDI32.SRCCOPY);
                //We select the old bitmap back to the memory device context.
                GDI32.SelectObject(hMemDC, hOld);
                //We delete the memory device context.
                GDI32.DeleteDC(hMemDC);
                //We release the screen device context.
                USER32.ReleaseDC(USER32.GetDesktopWindow(), hDC);
                //Image is created by Image bitmap handle and stored in local variable.
                Bitmap bmp = System.Drawing.Image.FromHbitmap(hBitmap);
                //Release the memory for compatible bitmap.
                GDI32.DeleteObject(hBitmap);
                //This statement runs the garbage collector manually.
                GC.Collect();
                //Return the bitmap 
                return bmp;
            }

            //If hBitmap is null retunrn null.
            return null;
        }

        public static Bitmap GetWindowImage(IntPtr hWnd, WINDOWTYPE type)
        {
            switch (type)
            {
                case WINDOWTYPE.PW_WINDOW: // copy entire window 
                    {
                        USER32.RECT rectWnd = new USER32.RECT(0, 0, 0, 0);

                        // get the window rectangle 

                        USER32.GetWindowRect(hWnd, ref rectWnd);

                        // get the bitmap of that window by calling 
                        // CopyScreenToBitmap and passing it the window rect 

                        return GetScreenImage(rectWnd);
                    }


                case WINDOWTYPE.PW_CLIENT: // copy client area 
                    {
                        USER32.RECT rectClient = new USER32.RECT(0, 0, 0, 0) ;
                        USER32.POINT pt1, pt2;

                        // get client dimensions 

                        USER32.GetClientRect(hWnd, ref rectClient);

                        // convert client coords to screen coords 

                        pt1.x = rectClient.left;
                        pt1.y = rectClient.top;
                        pt2.x = rectClient.right;
                        pt2.y = rectClient.bottom;
                        USER32.ClientToScreen(hWnd, ref pt1);
                        USER32.ClientToScreen(hWnd, ref pt2);
                        rectClient.left = pt1.x;
                        rectClient.top = pt1.y;
                        rectClient.right = pt2.x;
                        rectClient.bottom = pt2.y;

                        // get the bitmap of the client area by calling 
                        // CopyScreenToBitmap and passing it the client rect 

                        return GetScreenImage(rectClient);
                    }

                default:    // invalid print area 
                    return null;
            }
        }

        public static Bitmap GetClientImage(IntPtr hWnd)
        {
            return GetWindowImage(hWnd, WINDOWTYPE.PW_CLIENT);
        }

		#endregion
	}
}
