/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;

namespace Elearning.MySocket
{
    public class TCPClient
    {
        Socket m_sock;

        private AsyncCallback pfnReceivingCallBack;
        public delegate void ReceivedMsgEventHandler(object sender, TCPSocketEventArgs e);
        public event ReceivedMsgEventHandler ReceivedMessage;

        public TCPClient(String strIP, int port)
        {
            IPAddress ip;

            /**********************obsoleted
             * if (strIP[0] >= '0' && strIP[0] <= '9')
             * ip = IPAddress.Parse(strIP);
             * else
             * ip = Dns.GetHostByName(strIP).AddressList[0];
             * */

            ip = Dns.GetHostEntry(strIP).AddressList[0];
            m_sock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            // Create the end point 
            IPEndPoint m_remoteEndPoint = new IPEndPoint(ip, port);

            // Connect to the remote host
            m_sock.Connect(m_remoteEndPoint);
        }

        public TCPClient(IPAddress ip, int port)
        {
            m_sock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            // Create the end point 
            IPEndPoint m_remoteEndPoint = new IPEndPoint(ip, port);

            // Connect to the remote host
            m_sock.Connect(m_remoteEndPoint);
        }

        public void Start()
        {
            if (m_sock.Connected)
            {
                // first message to notify the client that it was connected to the server
                ReceivedMessage(this, new TCPSocketEventArgs(m_sock, 0, null));

                //Wait for data asynchronously 
                StateObjectTcp sockPkt = new StateObjectTcp();
                sockPkt.sock = m_sock;
                sockPkt.seq = 0;

                WaitForTCPData(sockPkt);
            }
        }

        public void Close()
        {
            if (m_sock != null)
            {
                //m_sock.Disconnect(true);
                m_sock.Close();
                m_sock = null;
            }
        }

        public void WaitForTCPData(StateObjectTcp sockPkt)
        {
            try
            {
                if (pfnReceivingCallBack == null)
                {
                    pfnReceivingCallBack = new AsyncCallback(OnTCPDataReceived);
                }

                // Start receiving any data written by the connected client asynchronously
                sockPkt.sock.BeginReceive(sockPkt.dataBuffer,
                    0,
                    StateObjectTcp.BUFFER_SIZE,
                    SocketFlags.None,
                    pfnReceivingCallBack,
                    sockPkt
                    );
            }
            catch (SocketException se)
            {
                MessageBox.Show(se.Message);
            }
        }

        public void OnTCPDataReceived(IAsyncResult asyn)
        {
            StateObjectTcp socketPkt = (StateObjectTcp)asyn.AsyncState;
            try
            {
                socketPkt.seq++;

                int iRx = socketPkt.sock.EndReceive(asyn);
                byte[] bdata = new byte[iRx];

                for (int i = 0; i < iRx; i++)
                {
                    bdata[i] = socketPkt.dataBuffer[i];
                }
                ReceivedMessage(this, new TCPSocketEventArgs(socketPkt.sock, socketPkt.seq, bdata));

                WaitForTCPData(socketPkt);
            }
            catch (ObjectDisposedException)
            {
                System.Diagnostics.Debugger.Log(0, "1", "\nOnDataReceived: Socket has been closed\n");
            }
            catch (SocketException se)
            {
                if (se.ErrorCode == 10054)
                    ReceivedMessage(this, new TCPSocketEventArgs(socketPkt.sock, -1, null));
                MessageBox.Show(se.Message);
            }
        }

        public void Send(byte[] data)
        {
            m_sock.Send(data);
        }

        public bool Connected
        {
            get
            {
                return m_sock.Connected;
            }
        }

    }
}
