/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.Collections;

namespace Elearning.MySocket
{
    public class StateObjectTcp
    {
        public Socket sock = null;
        public long seq = 0;
        public const int BUFFER_SIZE = 1024;
        public byte[] dataBuffer = new byte[BUFFER_SIZE];
    }


    public class TCPServer
    {
        public const int PORTCHBOX = 5367, MAX_CLIENT_NUMBER = 50;
        private Socket m_sock;
        private int m_nPort;

        private AsyncCallback pfnReceivingCallBack;
        public delegate void ReceivedMsgEventHandler(object sender, TCPSocketEventArgs e);
        public event ReceivedMsgEventHandler ReceivedMessage;


        public TCPServer()
        {
            m_nPort = PORTCHBOX;
        }

        public TCPServer(int port)
        {
            m_nPort = port;
        }

        public void Close()
        {
            if (m_sock != null)
            {
                // m_sockChbox.Disconnect(true);
                m_sock.Close();
                m_sock = null;
            }
        }

        public void Start()
        {
            //create a new client socket ...
            m_sock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPEndPoint ipPoint = new IPEndPoint(IPAddress.Any, m_nPort);

            m_sock.Bind(ipPoint);
            m_sock.Listen(MAX_CLIENT_NUMBER);
            m_sock.BeginAccept(new AsyncCallback(OnTCPClientConnect), null);
        }

        public void OnTCPClientConnect(IAsyncResult asyn)
        {
            try
            {
                StateObjectTcp sockPkt = new StateObjectTcp();
                sockPkt.sock = m_sock.EndAccept(asyn);
                sockPkt.seq = 0;
                // seq = 0 means this is a connection, instead of message
                ReceivedMessage(this, new TCPSocketEventArgs(sockPkt.sock, 0, null));

                WaitForTCPData(sockPkt);
                m_sock.BeginAccept(new AsyncCallback(OnTCPClientConnect), null);
            }
            catch (ObjectDisposedException)
            {
                System.Diagnostics.Debugger.Log(0, "1", "\n OnTCPClientConnection: Socket has been closed\n");
            }
            catch (SocketException se)
            {
                MessageBox.Show(se.Message);
            }
        }

        public void WaitForTCPData(StateObjectTcp tcpPkt)
        {
            try
            {
                if (pfnReceivingCallBack == null)
                {
                    // Specify the call back function which is to be 
                    // invoked when there is any write activity by the 
                    // connected client
                    pfnReceivingCallBack = new AsyncCallback(OnTCPDataReceived);
                }

                // Start receiving any data written by the connected client asynchronously
                tcpPkt.sock.BeginReceive(tcpPkt.dataBuffer,
                    0,
                    StateObjectTcp.BUFFER_SIZE,
                    SocketFlags.None,
                    pfnReceivingCallBack,
                    tcpPkt
                    );
            }
            catch (SocketException se)
            {
                MessageBox.Show(se.Message);
            }
        }

        public void OnTCPDataReceived(IAsyncResult asyn)        
        {
            StateObjectTcp socketPkt = (StateObjectTcp)asyn.AsyncState;
            try
            {                
                socketPkt.seq ++;

                // Complete the BeginReceive() asynchronous call by EndReceive() method
                // which will return the number of characters written to the stream 
                // by the client
                int iRx = socketPkt.sock.EndReceive(asyn);
                byte[] bdata = new byte[iRx];

                for (int i = 0; i < iRx; i++)
                {
                    bdata[i] = socketPkt.dataBuffer[i];
                }
                ReceivedMessage(this, new TCPSocketEventArgs(socketPkt.sock, socketPkt.seq, bdata));

                WaitForTCPData(socketPkt);
            }
            catch (ObjectDisposedException)
            {
                System.Diagnostics.Debugger.Log(0, "1", "\nOnDataReceived: Socket has been closed\n");
            }
            catch (SocketException se)
            {
                if (se.ErrorCode == 10054)
                    ReceivedMessage(this, new TCPSocketEventArgs(socketPkt.sock, -1, null));
                MessageBox.Show(se.Message);
            }
        }
    }

    public class TCPSocketEventArgs : EventArgs
    {        
        private byte[] msg;
        private Socket sock;
        private long seq;
        public TCPSocketEventArgs(Socket s, long n, byte[] m)
        {
            sock = s;
            seq = n;
            msg = m;
        }
        public Socket m_sock
        {
            get
            {
                return sock;
            }
        }
        public long m_seq
        {
            get
            {
                return seq;
            }
        }
        public byte[] m_msg
        {
            get
            {
                return msg;
            }
        }
    }

}
