/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using Elearning.VideoCap;
using Elearning.MySocket;

namespace Elearning.Student
{
    public partial class FormStudent : Form
    {
        private String m_strTeacherName, m_strStudentName;

        private TCPClient c_chbox;
        private UDPServer c_video, c_slide;
        private Socket m_sockSend;
        private IPEndPoint m_audioEndPoint, m_videoEndPoint;
        private WebCamera wc;
        private MassWare.MassWare m_massWare;

        public delegate void UpdateTextCallback(string text);
        public delegate void UpdateVideoCallback(Image im);

        #region system functions
        public FormStudent()
        {
            InitializeComponent();
            UpdateControls(false);
        }

        private void FormStudent_Load(object sender, EventArgs e)
        {
            wc = new WebCamera(picBoxSelf.Handle, picBoxSelf.Width, picBoxSelf.Height);
            wc.RecievedFrame += new WebCamera.RecievedFrameEventHandler(wc_RecievedFrame);
            m_massWare = new MassWare.MassWare(@"D:\MassWare\Outputs\elRules.xml");
            m_massWare.Initialize();
            object[] paramlist = new object[3] { picBoxVideo.Handle, picBoxVideo.Width, picBoxVideo.Height };
            m_massWare.SetAppParameter("Masslets.ImageProcess.Display", "SetDisplayInf", paramlist);

        }

        private void FormStudent_FormClosing(object sender, FormClosingEventArgs e)
        {
            timerProc.Stop();
            wc.CloseWebcam();
            CloseSockets();
        }

        private void connectToToolStripMenuItem_Click(object sender, EventArgs e)
        {
            FormConnectTo connectForm = new FormConnectTo();
            if (connectForm.ShowDialog(this) == DialogResult.OK)
            {
                m_strStudentName = connectForm.strName;
                try
                {
                    // get local ip address
                    // Dns.GetHostEntry(connectForm.strIP).AddressList[0];
                    IPAddress remoteIPAddress = IPAddress.Parse(connectForm.strIP);

                    // set the send socket for sending video and audio
                    m_sockSend = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                    // set the remote port number, provided by the server
                    m_videoEndPoint = new IPEndPoint(remoteIPAddress, connectForm.nPortVideo);

                    /////////////start the chat box function
                    c_chbox = new TCPClient(connectForm.strIP, connectForm.nPortChBox);
                    c_chbox.ReceivedMessage += new TCPClient.ReceivedMsgEventHandler(c_chBox_RecievedMessage);
                    c_chbox.Start();

                    ////////////start the video receiver
                    c_video = new UDPServer(UDPServer.PORTVIDEO);
                    c_video.ReceivedMessage += new UDPServer.ReceivedMsgEventHandler(c_video_ReceivedMessage);
                    c_video.Start();

                    ////////////start the slides receiver
                    c_slide = new UDPServer(UDPServer.PORTSLIDE);
                    c_slide.ReceivedMessage += new UDPServer.ReceivedMsgEventHandler(c_slide_ReceivedMessage);
                    c_slide.Start();

                    wc.StartWebCam(WebCamera.GrabType.tFrame);
                    timerProc.Interval = 66;
                    timerProc.Start();

                    if (c_chbox.Connected) UpdateControls(true);

                }
                catch (SocketException se)
                {
                    string str;
                    str = "\nConnection failed, is the server running?\n" + se.Message;
                    MessageBox.Show(str);
                    UpdateControls(false);
                } 
             }
            connectForm.Dispose();             
        }

        private void disconnectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            FormStudent_FormClosing(sender, null);
            UpdateControls(false);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void buttonSend_Click(object sender, EventArgs e)
        {
            if (textBoxSend.Text == "") return;
            try
            {
                textBoxSend.AppendText("\n");
                byte[] byData = System.Text.Encoding.ASCII.GetBytes(textBoxSend.Text);
                if (c_chbox != null) c_chbox.Send(byData);

                AppendText("You: " + textBoxSend.Text);
                textBoxSend.Text = "";
            }
            catch (SocketException se)
            {
                MessageBox.Show(se.Message);
            }
        }

        private void textBoxSend_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Return)
            {
                buttonSend_Click(sender, e);
                e.Handled = true;
            }
        }

        private void timerProc_Tick(object sender, EventArgs e)
        {
            if (c_chbox.Connected)
            {
                wc.GetOneFrame();
            }
        }
        #endregion

        #region self defined functions
        private void UpdateControls(bool connected)
        {
            connectToToolStripMenuItem.Enabled = !connected;
            disconnectToolStripMenuItem.Enabled = connected;
            string connectStatus = connected ? "Connected" : "Not Connected";
            statusStrip1.Text = connectStatus;
        }

        private void CloseSockets()
        {
            if (c_chbox != null) c_chbox.Close();
            if (c_video != null) c_video.Close();
            if (c_slide != null) c_slide.Close();
            if (m_sockSend != null)
            {
                m_sockSend.Close();
                m_sockSend = null;
            }
        }

        #endregion

        #region event response callback functions
        private void wc_RecievedFrame(object sender, WebCameraEventArgs e)
        {
            Image mImage = e.mImage.GetThumbnailImage(picBoxSelf.Width, picBoxSelf.Height, null, System.IntPtr.Zero);
            picBoxSelf.Image = mImage;

            byte[] byteImage = ImageConv.imageToByteArray(mImage, System.Drawing.Imaging.ImageFormat.Jpeg);
            int nBytesSent = m_sockSend.SendTo(byteImage, m_videoEndPoint);
        }

        private void c_chBox_RecievedMessage(object sender, TCPSocketEventArgs e)
        {
            String strDisp;

            switch (e.m_seq)
            {
                case -1: // connection closed forcely by remote server
                    {
                        disconnectToolStripMenuItem_Click(null, null);
                        return;
                    }
                case 0: // connected to the server, not message
                    {
                        // Send first massage, the name of the student
                        byte[] byData = System.Text.Encoding.ASCII.GetBytes(m_strStudentName);
                        c_chbox.Send(byData);
                        return;
                    }
                case 1: // received the first message
                    {
                        // get the name of the teacher
                        m_strTeacherName = System.Text.Encoding.ASCII.GetString(e.m_msg);

                        // Display this server connection as a status message on the GUI	
                        strDisp = String.Format("You have been connected to Teacher : {0}\r\n", m_strTeacherName);

                        break;
                    }
                default: // the rest message
                    {
                        strDisp = m_strTeacherName + ": " + System.Text.Encoding.ASCII.GetString(e.m_msg);
                        break;
                    }
            }

            // display to richtextbox
            richTextBoxMemo.Invoke(new UpdateTextCallback(AppendText), new object[] { strDisp });

        }

        private void c_video_ReceivedMessage(object sender, UDPSocketEventArgs e)
        {
            Image remoteImage = ImageConv.byteArrayToImage(e.m_data);
            picBoxVideo.Invoke(new UpdateVideoCallback(DisplayRemoteVideo), new object[] { remoteImage });
        }

        private void c_slide_ReceivedMessage(object sender, UDPSocketEventArgs e)
        {
            Image remoteImage = ImageConv.byteArrayToImage(e.m_data);
            picBoxSlide.Invoke(new UpdateVideoCallback(DisplayREmoteSlide), new object[] { remoteImage });
        }

        #endregion
    }
}