/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.Collections;
using Elearning.VideoCap;
using Elearning.MySocket;
using MassWare;

namespace Elearning.Teacher
{
    public partial class FormTeacher : Form
    {
        #region vairables
        private const int SLIDEWIDTH = 640, SLIDEHEIGHT = 480;
        private TCPServer c_chBox;
        private MySocket.UDPServer c_video;
        private String m_strTeacherName;
        private StudentTable m_studentTable;
        private MassWare.MassWare m_massWare;
        public int m_sendInterval, m_imageQuality;

        public delegate void UpdateTextCallback(string text);
        public delegate void UpdatePicBoxCallback(string tag);
        public delegate void UpdateVideoCallback(Image im, string tag);
        
        FormSlide formSlide;
        #endregion

        #region system functions
        public FormTeacher()
        {
            InitializeComponent();
            UpdateControls(false);
            m_sendInterval = 2000;
            m_imageQuality = 70;
        }

        private void FormTeacher_Load(object sender, EventArgs e)
        {
            m_studentTable = new StudentTable(TCPServer.MAX_CLIENT_NUMBER);

            formSlide = new FormSlide();
            formSlide.Show();
            hideToolStripMenuItem.Checked = formSlide.Visible;

            m_massWare = new MassWare.MassWare(@"D:\MassWare\Outputs\elRules.xml");
            m_massWare.Initialize();
            object[] paramlist = { picBoxAll.Handle, picBoxAll.Width, picBoxAll.Height };
            m_massWare.SetAppParameter("Masslets.ImageProcess.WebCam", "SetLocalWindowInf", paramlist);
        }

        private void startToolStripMenuItem_Click(object sender, EventArgs e)
        {

            FormStart startForm = new FormStart();
            if (startForm.ShowDialog(this) == DialogResult.OK)
            {
                m_strTeacherName = startForm.strName;
                try
                {
                    ///////////////////////////////////////////////////////////////////////////////chatbox
                    c_chBox = new TCPServer(startForm.portChbox);
                    c_chBox.ReceivedMessage += new TCPServer.ReceivedMsgEventHandler(c_chBox_RecievedMessage);
                    c_chBox.Start();
                    
                    ////////////////////////////////////////////////////////////////////////////////video
                    c_video = new MySocket.UDPServer(startForm.portVideo);
                    c_video.ReceivedMessage += new MySocket.UDPServer.ReceivedMsgEventHandler(c_video_ReceivedMessage);
                    c_video.Start();

                    timerProc.Interval = m_sendInterval;
                    timerProc.Start();

                    m_massWare.Start();
                    UpdateControls(true);
                }
                catch (SocketException se)
                {
                    MessageBox.Show(se.Message);
                    UpdateControls(false);
                }
            }
            startForm.Dispose();
        }

        private void stopToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CloseSockets();
            m_massWare.Stop();
            UpdateControls(false);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void hideToolStripMenuItem_Click(object sender, EventArgs e)
        {
            formSlide.Visible = !formSlide.Visible;
            hideToolStripMenuItem.Checked = formSlide.Visible;
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            FormPref formPref = new FormPref();
            formPref.m_interval = m_sendInterval;
            formPref.m_quality = m_imageQuality;
            if (formPref.ShowDialog(this) == DialogResult.OK)
            {
                m_imageQuality = formPref.m_quality;
                m_sendInterval = formPref.m_interval;
            }
            formPref.Dispose();

            if (timerProc.Enabled && m_sendInterval != timerProc.Interval)
            {
                timerProc.Stop();
                timerProc.Interval = m_sendInterval;
                timerProc.Start();
            }
        }

        private void FormTeacher_FormClosing(object sender, FormClosingEventArgs e)
        {
            stopToolStripMenuItem_Click(sender, null);
            m_massWare.Dispose();
            formSlide.Close();
            GC.Collect();
        }

        private void buttonSend_Click(object sender, EventArgs e)
        {
            if (textBoxSend.Text == "") return;
            try
            {
                textBoxSend.AppendText("\r\n");
                byte[] byData = System.Text.Encoding.ASCII.GetBytes(textBoxSend.Text);

                m_studentTable.SendMsg(byData);

                AppendText("You: " + textBoxSend.Text);
                textBoxSend.Text = "";

            }
            catch (SocketException se)
            {
                MessageBox.Show(se.Message);
            }
        }

        private void textBoxSend_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Return)
            {
                buttonSend_Click(sender, e);
                e.Handled = true;
            }
        }

        private void picBoxAll_Click(object sender, EventArgs e)
        {
            if ((String)(((PictureBox)sender).Tag) == "All") picBoxAll.BorderStyle = BorderStyle.Fixed3D;
            else picBoxAll.BorderStyle = BorderStyle.FixedSingle;
            m_studentTable.m_strActiveClientIP = (String)((PictureBox)sender).Tag;
        }

        private void timerProc_Tick(object sender, EventArgs e)
        {
            if (formSlide.Visible)
            {
                Image im = ImageProc.CaptureWindow.GetClientImage(formSlide.Handle);
                Image imStretch = ImageConv.StretchImage(im, SLIDEWIDTH, SLIDEHEIGHT, true);
                byte[] imdata = ImageConv.imageToJpegByteArray(imStretch, m_imageQuality);
                m_studentTable.SendSlide(imdata);
            }
        }

        #endregion

        #region self defined functions
        private void CloseSockets()
        {
            if (c_chBox != null) c_chBox.Close();
            if (c_video != null) c_video.Close();

            if (m_studentTable != null) m_studentTable.Close();
        }

        private void UpdateControls(bool connected)
        {
            startToolStripMenuItem.Enabled = !connected;
            stopToolStripMenuItem.Enabled = connected;
            string connectStatus = connected ? "Connected" : "Not Connected";
            statusStrip1.Text = connectStatus;
        }

        #endregion

        #region event response callback functions
        private void c_chBox_RecievedMessage(object sender, TCPSocketEventArgs e)
        {
            String strDisp;
            IPAddress remoteIP = ((IPEndPoint)(e.m_sock.RemoteEndPoint)).Address;
            String strRemoteIP = remoteIP.ToString();

            switch (e.m_seq)
            {
                case -1: // connection closed by remote socket
                    {
                        m_studentTable[strRemoteIP].m_picBox.Invoke(new UpdatePicBoxCallback(DisposeAPicBox), new object[] { strRemoteIP });
                        m_studentTable.Remove(strRemoteIP);
                        return;
                    }
                case 0: // received a connection, not message
                    {
                        StudentRecord record = new StudentRecord();
                        record.m_sockChbox = e.m_sock;
                        record.m_endPointVideo = new IPEndPoint(remoteIP, MySocket.UDPServer.PORTVIDEO);
                        record.m_endPointAudio = new IPEndPoint(remoteIP, MySocket.UDPServer.PORTAUDIO);
                        record.m_endPointSlide = new IPEndPoint(remoteIP, MySocket.UDPServer.PORTSLIDE);
                        m_studentTable.Add(strRemoteIP, record);
                        m_massWare.SyncTo(remoteIP);

                        // create a new picture box for this student
                        panelVideo.Invoke(new UpdatePicBoxCallback(CreateANewPicBox), new object[] { strRemoteIP });

                        // send the teacher's name to the student
                        e.m_sock.Send(System.Text.Encoding.ASCII.GetBytes(m_strTeacherName));
                        return;
                    }
                case 1: // the first message, it is the name of the student
                    {
                        String strName = System.Text.Encoding.ASCII.GetString(e.m_msg);

                        /* another way to change byte[] to string
                         * char[] chars = new char[iRx + 1];
                         * System.Text.Decoder dec = System.Text.Encoding.UTF8.GetDecoder();
                         * int charLen = dec.GetChars(socketPkt.dataBuffer, 0, iRx, chars, 0);
                         * String szData = new System.String(chars);
                         * */

                        m_studentTable[strRemoteIP].m_strName = strName;

                        // display to richtextbox
                        strDisp = String.Format("Student {0} from IP: {1} connected\r\n", strName, strRemoteIP);
                        break;
                    }
                default:
                    {
                        strDisp = m_studentTable[strRemoteIP].m_strName + ": " + System.Text.Encoding.ASCII.GetString(e.m_msg);
                        break;
                    }
            }

            // display to richtextbox
            richTextBoxMemo.Invoke(new UpdateTextCallback(AppendText), new object[] { strDisp });

        }

        private void c_video_ReceivedMessage(object sender, UDPSocketEventArgs e)
        {
            Image remoteImage = ImageConv.byteArrayToImage(e.m_data);
            if (m_studentTable.ContainsKey(e.m_strIp))
                m_studentTable[e.m_strIp].m_picBox.Invoke(new UpdateVideoCallback(DisplayRemoteVideo), new object[] { remoteImage, e.m_strIp });
        }

        #endregion
    }
}