/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Windows.Forms;
using System.Drawing;
using System.Runtime.InteropServices;

namespace Elearning.VideoCap
{
    ///
    /// avicap ժҪ˵
    ///
    public class API32
    {
        // showVideo calls
        [DllImport("avicap32.dll")]
        public static extern IntPtr capCreateCaptureWindowA(byte[] lpszWindowName, int dwStyle, int x, int y, int nWidth, int nHeight, IntPtr hWndParent, int nID);
        [DllImport("avicap32.dll")]
        public static extern bool capGetDriverDescriptionA(short wDriver, byte[] lpszName, int cbName, byte[] lpszVer, int cbVer);
        [DllImport("avicap32.dll")]
        public static extern int capGetVideoFormat(IntPtr hWnd, IntPtr psVideoFormat, int wSize);
        [DllImport("User32.dll")]
        public static extern bool SendMessage(IntPtr hWnd, int wMsg, bool wParam, int lParam);
        [DllImport("User32.dll")]
        public static extern bool SendMessage(IntPtr hWnd, int wMsg, short wParam, int lParam);
        [DllImport("User32.dll")]
        public static extern bool SendMessage(IntPtr hWnd, int wMsg, short wParam, FrameEventHandler lParam);
        [DllImport("User32.dll")]
        public static extern bool SendMessage(IntPtr hWnd, int wMsg, int wParam, ref BITMAPINFO lParam);
        [DllImport("User32.dll")]
        public static extern int SetWindowPos(IntPtr hWnd, int hWndInsertAfter, int x, int y, int cx, int cy, int wFlags);

        // Constants
        public const int WM_USER = 0x400;
        public const int WS_CHILD = 0x40000000;
        public const int WS_VISIBLE = 0x10000000;
        public const int SWP_NOMOVE = 0x2;
        public const int SWP_NOZORDER = 0x4;
        public const int WM_CAP_START = WM_USER;
        public const int WM_CAP_DRIVER_CONNECT = WM_CAP_START + 10;
        public const int WM_CAP_DRIVER_DISCONNECT = WM_CAP_START + 11;
        public const int WM_CAP_SET_CALLBACK_FRAME = WM_CAP_START + 5;
        public const int WM_CAP_SET_PREVIEW = WM_CAP_START + 50;
        public const int WM_CAP_SET_PREVIEWRATE = WM_CAP_START + 52;
        public const int WM_CAP_SET_SCALE = WM_CAP_START + 53;
        public const int WM_CAP_GET_VIDEOFORMAT = WM_CAP_START + 44;
        public const int WM_CAP_SET_VIDEOFORMAT = WM_CAP_START + 45;
        public const int WM_CAP_GRAB_FRAME = WM_CAP_START + 60;

        // Structures
        [StructLayout(LayoutKind.Sequential)]
        public struct VIDEOHDR
        {
            [MarshalAs(UnmanagedType.I4)]
            public int lpData;
            [MarshalAs(UnmanagedType.I4)]
            public int dwBufferLength;
            [MarshalAs(UnmanagedType.I4)]
            public int dwBytesUsed;
            [MarshalAs(UnmanagedType.I4)]
            public int dwTimeCaptured;
            [MarshalAs(UnmanagedType.I4)]
            public int dwUser;
            [MarshalAs(UnmanagedType.I4)]
            public int dwFlags;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
            public int[] dwReserved;
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct BITMAPINFOHEADER
        {
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biSize;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biWidth;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biHeight;
            [MarshalAs(UnmanagedType.I2)]
            public short biPlanes;
            [MarshalAs(UnmanagedType.I2)]
            public short biBitCount;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biCompression;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biSizeImage;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biXPelsPerMeter;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biYPelsPerMeter;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biClrUsed;
            [MarshalAs(UnmanagedType.I4)]
            public Int32 biClrImportant;
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct BITMAPINFO
        {
            [MarshalAs(UnmanagedType.Struct, SizeConst = 40)]
            public BITMAPINFOHEADER bmiHeader;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 1024)]
            public Int32[] bmiColors;
        }

        public delegate void FrameEventHandler(IntPtr lwnd, IntPtr lpVHdr);

        // Public methods
        public static object GetStructure(IntPtr ptr, ValueType structure)
        {
            return Marshal.PtrToStructure(ptr, structure.GetType());
        }

        public static object GetStructure(int ptr, ValueType structure)
        {
            return GetStructure(new IntPtr(ptr), structure);
        }

        public static void Copy(IntPtr ptr, byte[] data)
        {
            Marshal.Copy(ptr, data, 0, data.Length);
        }

        public static void Copy(int ptr, byte[] data)
        {
            Copy(new IntPtr(ptr), data);
        }

        public static int SizeOf(object structure)
        {
            return Marshal.SizeOf(structure);
        }

        static byte[] StructureToByteArray(object obj)
        {
            int len = Marshal.SizeOf(obj);
            byte[] arr = new byte[len];

            IntPtr ptr = Marshal.AllocHGlobal(len);
            Marshal.StructureToPtr(obj, ptr, true);
            Marshal.Copy(ptr, arr, 0, len);
            Marshal.FreeHGlobal(ptr);

            return arr;
        }

        static void ByteArrayToStructure(byte[] bytearray, ref object obj)
        {
            int len = Marshal.SizeOf(obj);

            IntPtr i = Marshal.AllocHGlobal(len);
            Marshal.Copy(bytearray, 0, i, len);
            obj = Marshal.PtrToStructure(i, obj.GetType());
            Marshal.FreeHGlobal(i);
        }
    }

    //Web Camera Class
    public class WebCamera
    {
        public enum GrabType { tPreview = 0, tFrame }; 
        // Constructur
        public WebCamera(IntPtr handle, int width, int height)
        {
            mControlPtr = handle;
            mWidth = width;
            mHeight = height;
            mbStarted = false;
        }

        // constant value, support: 800*600, 640*480, 352*288, 320*240, 176*144, 160*120
        private const int VIDEOFRAMEWIDTH = 320;
        private const int VIDEOFRAMEHEIGHT = 240;

        // delegate for frame callback
        public delegate void RecievedFrameEventHandler(object sender, WebCameraEventArgs e);
        public event RecievedFrameEventHandler RecievedFrame;

        private IntPtr lwndC; // Holds the unmanaged handle of the control
        private IntPtr mControlPtr; // Holds the managed pointer of the control
        private int mWidth;
        private int mHeight;
        API32.BITMAPINFO mBitmapinfo;
        private bool mbStarted;

        private API32.FrameEventHandler mFrameEventHandler; // Delegate instance for the frame callback - must keep alive! gc should NOT collect it

        // Close the web camera
        public void CloseWebcam()
        {
            if (mbStarted)
            {
                this.capDriverDisconnect(this.lwndC);
                mbStarted = false;
            }
        }

        // start the web camera
        public bool StartWebCam(GrabType type)
        {
            byte[] lpszName = new byte[100];
            byte[] lpszVer = new byte[100];

            if (!API32.capGetDriverDescriptionA(0, lpszName, 100, lpszVer, 100)) return false;

            if (type == GrabType.tFrame)
            {  // this one is not efficent, just for click event response.
                this.lwndC = API32.capCreateCaptureWindowA(lpszName, 0, 0, 0, mWidth, mHeight, IntPtr.Zero, 0);
            }
            else if (type == GrabType.tPreview)
            {  // use this one if possible
                this.lwndC = API32.capCreateCaptureWindowA(lpszName, API32.WS_VISIBLE + API32.WS_CHILD, 0, 0, mWidth, mHeight, mControlPtr, 0);
            }

            if (this.lwndC == null) return false;

            if (this.capDriverConnect(this.lwndC, 0))
            {
                mBitmapinfo = new API32.BITMAPINFO();
                this.capGetVideoFormat(this.lwndC, ref mBitmapinfo, API32.SizeOf(mBitmapinfo));

                if (mBitmapinfo.bmiHeader.biWidth != VIDEOFRAMEWIDTH || mBitmapinfo.bmiHeader.biHeight != VIDEOFRAMEHEIGHT)
                {
                    mBitmapinfo.bmiHeader.biWidth = VIDEOFRAMEWIDTH;
                    mBitmapinfo.bmiHeader.biHeight = VIDEOFRAMEHEIGHT;
                    this.capSetVideoFormat(this.lwndC, ref mBitmapinfo, API32.SizeOf(mBitmapinfo));
                }

                this.mFrameEventHandler = new API32.FrameEventHandler(FrameCallBack);
                this.capSetCallbackOnFrame(this.lwndC, this.mFrameEventHandler);
                API32.SetWindowPos(this.lwndC, 0, 0, 0, mWidth, mHeight, 6);

                if (type == GrabType.tFrame)
                {
                    capGrabFrame(this.lwndC);
                }
                else if (type == GrabType.tPreview)
                {
                    this.capPreviewRate(this.lwndC, 66);
                    this.capPreviewScale(this.lwndC, true);
                    this.capPreview(this.lwndC, true);
                }
                
                mbStarted = true;
            }
            return true;
        }

        public void GetOneFrame()
        {
            if (mbStarted)
                capGrabFrame(this.lwndC);
        }

        // private functions
        private bool capDriverConnect(IntPtr lwnd, short i)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_DRIVER_CONNECT, i, 0);
        }

        private bool capDriverDisconnect(IntPtr lwnd)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_DRIVER_DISCONNECT, 0, 0);
        }

        private bool capPreview(IntPtr lwnd, bool f)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_SET_PREVIEW, f, 0);
        }

        private bool capPreviewRate(IntPtr lwnd, short wMS)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_SET_PREVIEWRATE, wMS, 0);
        }

        private bool capPreviewScale(IntPtr lwnd, bool f)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_SET_SCALE, f, 0);
        }

        private bool capGrabFrame(IntPtr lwnd)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_GRAB_FRAME, 0, 0);
        }

        private bool capSetCallbackOnFrame(IntPtr lwnd, API32.FrameEventHandler lpProc)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_SET_CALLBACK_FRAME, 0, lpProc);
        }

        private bool capGetVideoFormat(IntPtr lwnd, ref API32.BITMAPINFO BmpFormat, int CapFormatSize)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_GET_VIDEOFORMAT, CapFormatSize, ref BmpFormat);
        }

        private bool capSetVideoFormat(IntPtr lwnd, ref API32.BITMAPINFO BmpFormat, int CapFormatSize)
        {
            return API32.SendMessage(lwnd, API32.WM_CAP_SET_VIDEOFORMAT, CapFormatSize, ref BmpFormat);
        }

        private void FrameCallBack(IntPtr lwnd, IntPtr lpVHdr)
        {
            API32.VIDEOHDR videoHeader = new API32.VIDEOHDR();

            videoHeader = (API32.VIDEOHDR)API32.GetStructure(lpVHdr, videoHeader);
/*
            byte[] VideoData;
            VideoData = new byte[videoHeader.dwBytesUsed];
            API32.Copy(videoHeader.lpData, VideoData);
*/
            if (this.RecievedFrame != null)
            {
                WebCameraEventArgs e = new WebCameraEventArgs(null, (Bitmap)ImageConv.infoDataToImage(mBitmapinfo, videoHeader.lpData));
                this.RecievedFrame(this,e);
            }
        }

        public RecievedFrameEventHandler DataReady
        {
            get
            {
                return RecievedFrame;
            }
            set
            {
                RecievedFrame += value;
            }
        }
    }

    public class WebCameraEventArgs : EventArgs
    {
        private System.Drawing.Image image;
        private byte[] data;
        public WebCameraEventArgs(byte[] videoData, System.Drawing.Image img)
        {
            data = videoData;
            image = img;
        }
        public byte[] mData
        {
            get
            {
                return data;
            }
        }
        public System.Drawing.Image mImage
        {
            get
            {
                return image;
            }
        }
    }
}