/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.IO;
using System.Timers;

namespace MassTools.NetworkAwareness.AvailableBW
{
    class PTRClient
    {
        [DllImport("kernel32.dll")]
        extern static short QueryPerformanceCounter(ref long x);
        [DllImport("kernel32.dll")]
        extern static short QueryPerformanceFrequency(ref long x);

        string version = "2.1";

        int delay_num = 0;
        int packet_size = Common.PacketSize;
        int phase_num = 3;
        int probe_num = Common.ProbeNum;

        StreamWriter trace_fp = null; //FILE * trace_fp = NULL;
        bool verbose = false;
        bool debug = false;

        double b_bw = 0, competing_bw, PTR_bw, a_bw;
        double[] c_bw = new double[Common.MaxRepeat];
        int probing_phase_count = 0;
        long probing_start_time = 0, probing_end_time = 0;
        long tick_freq = 0;

        int dst_port = Common.START_PORT, probing_port;
        IPAddress dst_ip, src_ip;
        string dst_hostname, src_hostname; //char dst_hostname[MAXHOSTNAMELEN], src_hostname[MAXHOSTNAMELEN];
        string dst_ip_str, src_ip_str; //char dst_ip_str[16], src_ip_str[16];

        string dst, src; // char dst[MAXHOSTNAMELEN], src[MAXHOSTNAMELEN];
        Socket control_sock, probing_sock;

        IPEndPoint probing_server, probing_server2; //struct sockaddr_in probing_server, probing_server2;

        //RETSIGTYPE (*oldhandler)(int);

        double src_gap_sum = 0; // entry value
        double avg_src_gap; // entry value
        Common.pkt_rcd_t[] dst_gap = new Common.pkt_rcd_t[Common.MaxProbeNum];

        double dst_gap_sum;
        double avg_dst_gap;
        int dst_gap_count, src_gap_count;
        int total_count;
        long[] send_times = new long[Common.MaxProbeNum];
        double tlt_src_gap, tlt_dst_gap;

        // keep the data from dst machine, 128 is for other control string
        byte[] msg_buf = new byte[Common.pkt_rcd_t.Size() * Common.MaxProbeNum + 128]; // char msg_buf[sizeof(struct pkt_rcd_t) * MaxProbeNum + 128];
        int msg_len;

        // Shengpu added
        // create an event delegant
        public delegate void ABWReadyEventHandler(object sender, PTREventArgs e);
        // create an event to notify subscribers
        public event ABWReadyEventHandler ABWReadyEvent;

        // create a timer for frequently test
        // System.Timers.Timer sendingTimer = null;

        // trace item: record the item used to dump out into trace file
        class trace_item
        {
            public int probe_num, packet_size, delay_num;
            public long[] send_times = new long[Common.MaxProbeNum];
            public Common.pkt_rcd_t[] rcv_record = new Common.pkt_rcd_t[Common.MaxProbeNum];
            public int record_count;

            public double avg_src_gap;
            public double avg_dst_gap;

            public double b_bw, c_bw, a_bw, ptr;

            public trace_item next;

            public trace_item()
            {
                next = null;
            }
        };

        trace_item trace_list = null;
        trace_item trace_tail = null;
        trace_item cur_trace = null;

        public PTRClient(int probNum, int repeatNum, int packSize, string traceFile, string dstAddress, int dstPort, bool v, bool d, bool h)
        {
            QueryPerformanceFrequency(ref tick_freq);
            //sendingTimer = new System.Timers.Timer(5000);
            //sendingTimer.Elapsed += new ElapsedEventHandler(sendingTimer_Elapsed);


            if (probNum > Common.MaxProbeNum)
                probe_num = Common.MaxProbeNum;
            else
                probe_num = probNum;
            
            if (repeatNum > Common.MaxRepeat)
                phase_num = Common.MaxRepeat;
            else 
                phase_num = repeatNum;

            packet_size = packSize;
            trace_fp = File.CreateText(traceFile);

            dst = dstAddress;
            dst_port = dstPort;

            debug = d;
            verbose = v;
        }

        public PTRClient()
        {
            QueryPerformanceFrequency(ref tick_freq);

            //sendingTimer = new System.Timers.Timer(10000);
            //sendingTimer.Elapsed += new ElapsedEventHandler(sendingTimer_Elapsed);
            dst = "localhost";
        }

        public PTRClient(string dstAddress)
        {
            QueryPerformanceFrequency(ref tick_freq);

            //sendingTimer = new System.Timers.Timer(5000);
            //sendingTimer.Elapsed += new ElapsedEventHandler(sendingTimer_Elapsed);
            dst = dstAddress;
        }

        public void SetFrequency(int milliSec)
        {
            //sendingTimer.Interval = milliSec;
        }

        public void SetRemoteAddr(string strAddr)
        {
            dst = strAddr;
        }

        public bool Start()
        {
            //sendingTimer.Start();
            sendingTimer_Elapsed(this, null);
            return true;
        }

        public bool Stop()
        {
            //sendingTimer.Stop();
            return true;
        }

        private void sendingTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            init_connection();
            // allow dst 2 seconds to start up the packet filter
            System.Threading.Thread.Sleep(2000);

            // probing
            if (delay_num > 0)
            {
                one_phase_probing();
                get_bandwidth();
                dump_bandwidth();
            }
            else
                fast_probing();

            // finishing
            cleanup(1);

        }

        void Usage()
        {
            Console.WriteLine("IGI/PTR-{0} client usage:\n", version);
            Console.WriteLine("\tptr-client [-n probe_num] [-s packet_size] [-p dst_port]");
            Console.WriteLine("\t           [-k repeat_num] [-f trace_file] [-vdh] dst_address\n");
            Console.WriteLine("\t-n      set the number of probing packets in each train [60]");
            Console.WriteLine("\t-s      set the length of the probing packets in byte [500B]");
            Console.WriteLine("\t-p      indicate the dst machine's listening port [10241]");
            Console.WriteLine("\t        This is optional, it can itself search for the port");
            Console.WriteLine("\t        that the igi_server is using.");
            Console.WriteLine("\t-k      the number of train probed for each source gap [3]");
            Console.WriteLine("\t-f      dump packet-level trace into trace_file");
            Console.WriteLine("\t-v      verbose mode.");
            Console.WriteLine("\t-d      debug mode.");
            Console.WriteLine("\t-h      print this message.");
            Console.WriteLine("dst_address     can be either an IP address or a hostname\n");
        }

        /* combine the sec & usec of record into a real number */
        double get_rcd_time(Common.pkt_rcd_t record)
        {
            return (double)record.tick / tick_freq;
        }

        /* dump out all the trace packet time stamps */
        void dump_trace()
        {
            trace_item p;
            int i, index;

            if (trace_fp == null)
            {
                Console.WriteLine("-w not specified, no trace to dump");
                return;
            }

            // first dump out the summary data 
            p = trace_list;
            index = 0;
            trace_fp.WriteLine("\n%%probe_num packet_size delay_num avg_src_gap arv_dst_gap b_bw c_bw a_bw ptr");
            trace_fp.WriteLine("summary_data = [");
            while (p != null)
            {
                index++;
                trace_fp.WriteLine("%2d %4d %5d %f %f %12.3f %12.3f %12.3f %12.3f",
                        p.probe_num,
                        p.packet_size,
                        p.delay_num,
                        p.avg_src_gap,
                        p.avg_dst_gap,
                        p.b_bw,
                        p.c_bw,
                        p.a_bw,
                        p.ptr);
                p = p.next;
            }
            trace_fp.WriteLine("];");
            trace_fp.WriteLine("phase_num = {0}; \n", index);

            // then the detail packet trace
            p = trace_list;
            index = 0;
            while (p != null)
            {
                index++;

                trace_fp.WriteLine("send_time_{0} = [", index);
                for (i = 1; i < p.probe_num; i++)
                {
                    trace_fp.Write("{0} ", get_time_span(p.send_times[i - 1], p.send_times[i]));
                }
                trace_fp.WriteLine("];");
                trace_fp.WriteLine("send_array_size({0}) = {1}; \n", index, p.probe_num - 1);

                trace_fp.WriteLine("recv_time_{0} = [", index);

                for (i = 0; i < p.record_count - 1; i++)
                {
                    trace_fp.Write("{0} {1} ", get_rcd_time(p.rcv_record[i + 1]) - get_rcd_time(p.rcv_record[i]), p.rcv_record[i + 1].seq);
                }
                trace_fp.WriteLine("];");
                trace_fp.WriteLine("recv_array_size({0}) = {1}; \n", index, p.record_count - 1);

                p = p.next;
            }
        }

        /* make a clean exit on interrupts */
        void cleanup(int signo)
        {
            if (trace_fp != null)
            {
                dump_trace();
                trace_fp.Flush();
                trace_fp.Close();
            }

            probing_sock.Close();
            control_sock.Close();

            // free trace list 
            if (trace_list != null)
            {
                trace_list = null;
                GC.Collect();
            }
        }

        void quit()
        {
            trace_fp.Flush();
            cleanup(1);
            Environment.Exit(-1);
        }

        /* get the current time */
        long get_time()
        {
            long count = 0;
            QueryPerformanceCounter(ref count); // Begin timing.
            return count;
        }

        double get_time_span(long s_timetick, long e_timetick)
        {
            return (double)(e_timetick - s_timetick) / tick_freq;
        }

        /* find the delay number which can set the src_gap exactly as "gap" */
        int get_delay_num(double gap)
        {
            const int Scale = 10;
            int lower, upper, mid;
            long s_time, e_time;
            double tmp;
            int k;

            lower = 0;
            upper = 16;
            tmp = 133333;

            /* search for upper bound */
            s_time = e_time = 0;
            while (get_time_span(s_time, e_time) < gap * Scale)
            {
                s_time = get_time();
                for (k = 0; k < upper * Scale; k++)
                {
                    tmp = tmp * 7;
                    tmp = tmp / 13;
                }
                e_time = get_time();
                upper *= 2;
            }

            // binary search for delay_num 
            mid = (int)(upper + lower) / 2;
            while (upper - lower > 20)
            {
                s_time = get_time();
                for (k = 0; k < mid * Scale; k++)
                {
                    tmp = tmp * 7;
                    tmp = tmp / 13;
                }
                e_time = get_time();

                if (get_time_span(s_time, e_time) > gap * Scale)
                    upper = mid;
                else
                    lower = mid;
                mid = (int)(upper + lower) / 2;
            }

            return mid;
        }

        void dump_bandwidth()
        {
            Console.WriteLine("\nPTR: {0} Mpbs (suggested)", PTR_bw / 1000000);
            Console.WriteLine("IGI: {0} Mpbs", a_bw / 1000000);
            Console.WriteLine("Probing uses {0} seconds, {1} trains, ending at a gap value of {2} us.", get_time_span(probing_start_time, probing_end_time),
                probing_phase_count, (int)(tlt_dst_gap * 1000000));

            if (trace_fp != null)
            {
                trace_fp.WriteLine("%%Bottleneck Bandwidth: {0} Mbps", b_bw / 1000000);
                trace_fp.WriteLine("%%Competing  Bandwidth: {0} Mpbs", competing_bw / 1000000);
                trace_fp.WriteLine("%%Packet Transmit Rate: {0} Mpbs", PTR_bw / 1000000);
                trace_fp.WriteLine("%%Available  Bandwidth: {0} Mpbs", a_bw / 1000000);
            }

            ABWReadyEvent(this, new PTREventArgs(PTR_bw / 1000000, a_bw / 1000000));
        }

        int init_sockets(IPAddress dstIp)
        {
            IPEndPoint client = new IPEndPoint(IPAddress.Any, 0);

            probing_sock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);

            probing_server = new IPEndPoint(dstIp, probing_port);

            // used for sending the tail "junk" packet
            probing_server2 = new IPEndPoint(dstIp, probing_port + 1);

            try
            {
                probing_sock.Bind(client);
            }
            catch (SocketException e)
            {
                Console.WriteLine("Winsock error: " + e.ToString());
                quit();
            }
            return 1;
        }

        /* any parameter can be NULL */
        int get_host_info(string str, out string name, out IPAddress ip, out string ip_str)
        {
            if (str[0] >= '0' && str[0] <= '9')
            {
                ip = IPAddress.Parse(str);
                ip_str = str;
                name = Dns.GetHostByAddress(ip).HostName;
            }
            else
            {
                ip = Dns.GetHostByName(str).AddressList[0];
                ip_str = ip.ToString();
                name = str;
            }
            return 0;
        }

        /* the begining of the msg_buf should have the form $**$, return the string */
        // TODO: simplify using index()
        void get_item(Socket sock, out string str)
        {
            int i = 0, j, cur_len;
            bool bChecked = true;
            char[] msg_str = null;
            str = null;

            // check if two $ already exist
            if (msg_len > 0)
            {
                msg_str = Encoding.ASCII.GetChars(msg_buf, 0, msg_len);

                if (msg_str[0] != '$')
                {
                    Console.WriteLine("get unknown msg, it will mess up everything, exit");
                    quit();
                }
                i = 1;
                while ((i < msg_len) && (msg_str[i] != '$'))
                {
                    i++;
                }
                if (i < msg_len)
                {
                    string strtemp = new string(msg_str);
                    str = strtemp.Trim(new char[] { '$' });
                    bChecked = false;
                }
            }

            // read until we get two $
            while (bChecked)
            {
                if (debug)
                {
                    Console.Write("msg_len = {0}  ", msg_len);
                }
                cur_len = sock.Receive(msg_buf, msg_len, msg_buf.Length - msg_len, SocketFlags.None);

                // guard against control channel breaks
                if (cur_len <= 0)
                {
                    Console.WriteLine("read failed in get_item()");
                    quit();
                }

                msg_len += cur_len;
                msg_str = Encoding.ASCII.GetChars(msg_buf, 0, msg_len);

                i = 1;
                while ((i < msg_len) && (msg_str[i] != '$'))
                    i++;
                if (i < msg_len)
                {
                    string strtemp = new string(msg_str);
                    str = strtemp.Trim(new char[] { '$' });
                    break;
                }
            }

            // get out the string and move the buf ahead, with i points to the 2nd $
            i++;
            for (j = i; j < msg_len; j++)
                msg_buf[j - i] = msg_buf[j];
            msg_len -= i;
        }

        void init_connection()
        {
            string cur_str;
            string msg_str;

            // Get hostname and IP address of target host
            if (get_host_info(dst, out dst_hostname, out dst_ip, out dst_ip_str) < 0)
            {
                quit();
            }

            // search for the right dst port to connect with
            control_sock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            while (dst_port < Common.END_PORT)
            {
                try
                {
                    dst_port++;
                    if (verbose)
                        Console.WriteLine("dst_port = {0}", dst_port);
                    control_sock.Connect(dst_ip, dst_port);
                    break;
                }
                catch (SocketException e)
                {
                    Console.WriteLine("error: " + e.ToString());
                }
            }

            if (dst_port == Common.END_PORT)
            {
                Console.WriteLine("no dst port found, connection fails\n");
                quit();
            }

            // Get hostname and IP address of source host       
            src = Dns.GetHostName();
            if (get_host_info(src, out src_hostname, out src_ip, out src_ip_str) < 0)
            {
                Console.WriteLine("get source address info fails, exit");
                quit();
            }

            if (verbose)
            {
                Console.WriteLine("src addr: {0}", src_ip_str);
                Console.WriteLine("dst addr: {0}", dst_ip_str);
            }

            // send START message
            msg_str = "$START$" + probe_num + "$";
            byte[] sendingbuff = Encoding.ASCII.GetBytes(msg_str);
            control_sock.Send(sendingbuff);

            // wait for READY ack
            get_item(control_sock, out cur_str);
            if (verbose)
                Console.WriteLine("we get str: {0}", cur_str);
            if (!cur_str.StartsWith("READY"))
            {
                Console.WriteLine("get unknow str when waiting for READY from dst, exit");
                quit();
            }
            get_item(control_sock, out cur_str);

            if (verbose)
                Console.WriteLine("probing_port = {0}", cur_str);
            probing_port = int.Parse(cur_str);

            if (init_sockets(dst_ip) == 0)
            {
                Console.WriteLine("Error: init sockets failed");
                quit();
            }
        }

        /* send out probe_num packets, each packet is packet_size bytes, and the inital gap is set using delay_num */
        void send_packets(int probe_num, int packet_size, int delay_num, long[] sent_times)
        {
            int i, k;
            double tmp = 133333.0003333;

            byte[] send_buf = new byte[4096];

            // send out probing packets
            for (i = 0; i < probe_num - 1; i++)
            {
                // TODO: the middle send_times are not useful any more, since we don't use sanity-check
                sent_times[i] = get_time();
                send_buf[0] = (byte)i;
                probing_sock.SendTo(send_buf, packet_size, SocketFlags.None, probing_server);

                // gap generation
                for (k = 0; k < delay_num; k++)
                {
                    tmp = tmp * 7;
                    tmp = tmp / 13;
                }
            }

            // the last packets
            send_buf[0] = (byte)i;

            probing_sock.SendTo(send_buf, packet_size, SocketFlags.None, probing_server);

            probing_sock.SendTo(send_buf, 40, SocketFlags.None, probing_server2);
            sent_times[probe_num - 1] = get_time();
        }

        /* get dst gap_sum and gap_count from the records */
        double get_dst_sum(Common.pkt_rcd_t[] rcv_record, int count, out int gap_count)
        {
            double gap_sum, time1, time2;
            int i;

            gap_sum = 0;
            gap_count = 0;
            for (i = 0; i < count - 1; i++)
            {
                // only consider those gaps composed by two in-order packets 
                if (rcv_record[i + 1].seq - rcv_record[i].seq == 1)
                {
                    time1 = get_rcd_time(rcv_record[i + 1]);
                    time2 = get_rcd_time(rcv_record[i]);
                    gap_sum += (time1 - time2);
                    gap_count++;
                }
            }
            return gap_sum;
        }

        /* get dst gap trace data through control channel return 0 means fails in reading dst's data */
        int get_dst_gaps(Common.pkt_rcd_t[] rcv_record)
        {
            int i, cur_len, data_size, total_count, ptr;
            string num_str;

            get_item(control_sock, out num_str);
            data_size = int.Parse(num_str);
            total_count = data_size / Common.pkt_rcd_t.Size();
            if (verbose)
                Console.WriteLine("from dst: data_size = {0} total_count = {1} ", data_size, total_count);

            // get the data package
            while (msg_len < data_size)
            {
                cur_len = control_sock.Receive(msg_buf, msg_len, msg_buf.Length - msg_len, SocketFlags.None);
                if (cur_len <= 0)
                {
                    Console.WriteLine("read failed in get_dst_gaps()");
                    quit();
                }
                msg_len += cur_len;
            }

            // split out the record
            ptr = 0;
            byte[] tempdt = new byte[Common.pkt_rcd_t.Size()];
            for (i = 0; i < total_count; i++)
            {
                Array.Copy(msg_buf, ptr, tempdt, 0, Common.pkt_rcd_t.Size());
                ptr += Common.pkt_rcd_t.Size();
                rcv_record[i].FromByteArray(tempdt);

                if (debug)
                    Console.WriteLine("{0} {1} {2}", rcv_record[i].tick, tick_freq, rcv_record[i].seq);
            }
            msg_len -= data_size;

            if (trace_fp != null)
            {
                for (i = 0; i < total_count; i++)
                    cur_trace.rcv_record[i] = rcv_record[i];
                cur_trace.record_count = total_count;
            }

            return total_count;
        }
        
        struct bin_item 
        {
            public int value;
            public int count;
        };

        /* here we use some a similar technique with net timer, but the implementation is much simpler */
        double get_bottleneck_bw(Common.pkt_rcd_t[] rcv_record, int count)
        {
            double[] gaps = new double[Common.MaxProbeNum];
            int gap_count;
            bin_item[] bin = new bin_item[Common.MaxProbeNum];
            int bin_count;

            double lower_time, upper_time;
            int max_index, max_count, max_value, cur_value, i, j;

            double gap_sum;
            int sum_count;

            // get valid dst gaps
            gap_count = 0;
            for (i = 0; i < count - 1; i++)
            {
                if (rcv_record[i + 1].seq - rcv_record[i].seq == 1)
                {
                    gaps[gap_count] = get_rcd_time(rcv_record[i + 1]) - get_rcd_time(rcv_record[i]);
                    gap_count++;
                }
            }

            // use 25us bins to go through all the gaps
            bin_count = 0;
            for (i = 0; i < gap_count; i++)
            {
                cur_value = (int)(gaps[i] / Common.BinWidth);
                j = 0;
                while ((j < bin_count) && (cur_value != bin[j].value))
                {
                    j++;
                }
                if (j == bin_count)
                {
                    bin[bin_count].value = cur_value;
                    bin[bin_count].count = 1;
                    bin_count++;
                }
                else
                {
                    bin[j].count++;
                }
            }
            // find out the biggest bin
            max_index = 0; max_count = bin[0].count;
            for (i = 1; i < bin_count; i++)
            {
                if (bin[i].count > max_count)
                {
                    max_count = bin[i].count;
                    max_index = i;
                }
            }
            max_value = bin[max_index].value;
            lower_time = max_value * Common.BinWidth;
            upper_time = (max_value + 1) * Common.BinWidth;

            // see whether the adjacent two bins also have some items
            for (i = 0; i < bin_count; i++)
            {
                if (bin[i].value == max_value + 1)
                    upper_time += Common.BinWidth;
                if (bin[i].value == max_value - 1)
                    lower_time -= Common.BinWidth;
            }

            if (debug)
            {
                Console.WriteLine("get_bottleneck_bw: {0} {1}", lower_time, upper_time);
            }

            // average them
            gap_sum = 0;
            sum_count = 0;
            for (i = 0; i < gap_count; i++)
            {
                if ((gaps[i] >= lower_time) && (gaps[i] <= upper_time))
                {
                    gap_sum += gaps[i];
                    sum_count++;
                }
            }

            return (packet_size * 8 / (gap_sum / sum_count));
        }
        
        /* calculate the competing traffic rate using IGO method */
        double get_competing_bw(Common.pkt_rcd_t[] rcv_record, double avg_src_gap, int count, double b_bw)
        {
            double b_gap, cur_gap, m_gap;
            double gap_sum, inc_gap_sum;

            int i;

            b_gap = packet_size * 8 / b_bw;
            m_gap = Math.Max(b_gap, avg_src_gap);

            gap_sum = inc_gap_sum = 0;
            for (i = 0; i < count - 1; i++)
            {
                // only consider those gaps composed by two in-order packets
                if (rcv_record[i + 1].seq - rcv_record[i].seq == 1)
                {
                    cur_gap = get_rcd_time(rcv_record[i + 1]) - get_rcd_time(rcv_record[i]);
                    gap_sum += cur_gap;

                    if (cur_gap > m_gap + 0.000005)
                        inc_gap_sum += (cur_gap - b_gap);
                }
            }

            if (debug)
                Console.WriteLine("b_gap = {0} {1} {2} ", b_gap, inc_gap_sum, gap_sum);

            if (gap_sum == 0)
            {
                Console.WriteLine("\nNo valid trace in the last phase probing");
                Console.WriteLine("You may want to try again. Exit");
                quit();
            }

            return (inc_gap_sum * b_bw / gap_sum);
        }
        
        /* return summary of the "valid" src gaps */
        double get_src_sum(long[] times)
        {
            int i;
            double sum;
            
            sum = 0;
            for (i=0; i<probe_num-1; i++)
                sum += get_time_span(times[i], times[i + 1]);
            return sum;
        }

        void get_bandwidth()
        {
            if (dst_gap_count == 0)
                competing_bw = PTR_bw = a_bw = 0;
            else
            {
                if (b_bw < 1)
                    b_bw = get_bottleneck_bw(dst_gap, total_count);
                competing_bw = get_competing_bw(dst_gap, avg_src_gap, total_count, b_bw);
            }
        }
        
        /* one complete probing procedure, which includes:
         * 1. send out packets with (probe_num, packet_size, delay_num)
         * 2. retrieve the dst gaps values from dst machine
         * 3. filter out those bad trace data and compute the src gap & dst gap summary and average	
         * */
        void one_phase_probing()
        {
            if (verbose)
                Console.WriteLine("\nprobe_num = {0} packet_size = {1} delay_num = {2} ", probe_num, packet_size, delay_num);

            // probing
            send_packets(probe_num, packet_size, delay_num, send_times);
            if (trace_fp != null)
            {
                // create a new trace item
                cur_trace = new trace_item();
                cur_trace.next = null;
                if (trace_list == null)
                {
                    trace_list = cur_trace;
                    trace_tail = cur_trace;
                }
                else
                {
                    trace_tail.next = cur_trace;
                    trace_tail = cur_trace;
                }
            }
            // get dst gap logs
            total_count = get_dst_gaps(dst_gap);

            // compute avg_src_gap 
            src_gap_count = probe_num - 1;
            if (debug)
                Console.WriteLine("src_gap_count = {0} ", src_gap_count);

            src_gap_sum = get_src_sum(send_times);
            if (src_gap_count != 0)
                avg_src_gap = src_gap_sum / src_gap_count;
            else
                avg_src_gap = 0;

            // compute avg_dst_gap
            dst_gap_sum = get_dst_sum(dst_gap, total_count, out dst_gap_count);

            if (debug)
                Console.WriteLine("{0} dst_gap_count = {1} ", total_count, dst_gap_count);

            if (dst_gap_count != 0)
                avg_dst_gap = dst_gap_sum / dst_gap_count;
            else
                avg_dst_gap = 0;

            // record total src gaps
            tlt_src_gap += avg_src_gap;
            tlt_dst_gap += avg_dst_gap;

            if (verbose)
                Console.WriteLine("gaps (us): {0} {1} | {2} {3}",
                    tlt_src_gap * 1000000, tlt_dst_gap * 1000000,
                    avg_src_gap * 1000000, avg_dst_gap * 1000000);

            if (trace_fp != null)
            {
                double tmp1, tmp2, tmp3;

                cur_trace.probe_num = probe_num;
                cur_trace.packet_size = packet_size;
                cur_trace.delay_num = delay_num;

                for (int i = 0; i < probe_num; i++)
                {
                    cur_trace.send_times[i] = send_times[i];
                }
                cur_trace.avg_src_gap = avg_src_gap;
                cur_trace.avg_dst_gap = avg_dst_gap;

                // we can't use get_bandwidth() here since we want to record the current data, which may be wrong value
                if (dst_gap_count == 0)
                    cur_trace.b_bw = 0;
                else
                    cur_trace.b_bw = get_bottleneck_bw(dst_gap, total_count);

                tmp1 = PTR_bw;
                tmp2 = competing_bw;
                tmp3 = a_bw;

                get_bandwidth();
                cur_trace.ptr = PTR_bw;
                cur_trace.c_bw = competing_bw;
                cur_trace.a_bw = a_bw;
                PTR_bw = tmp1;
                competing_bw = tmp2;
                a_bw = tmp3;
            }

            probing_phase_count++;
        }

        void n_phase_probing(int n)
        {
            int i, j, k;

            tlt_src_gap = 0;
            tlt_dst_gap = 0;

            for (i = 0; i < n; i++)
            {
                one_phase_probing();
                get_bandwidth();

                // insert, so as to be able to pick the median easily
                j = 0;
                while (j < i && c_bw[j] < competing_bw)
                    j++;
                for (k = i - 1; k >= j; k--)
                    c_bw[k + 1] = c_bw[k];
                c_bw[j] = competing_bw;

                // Shengpu modified: to avoid the collision of receiving packages
                System.Threading.Thread.Sleep(300); 
            }
            PTR_bw = packet_size * 8 * phase_num / tlt_dst_gap;
            a_bw = b_bw - c_bw[(int)(n / 2)];

            if (verbose)
                Console.WriteLine("------------------------------------------------------");
        }

        int gap_comp(double dst_gap, double src_gap)
        {
            double delta = 0.05;

            if (dst_gap < src_gap / (1 + delta))
            {
                if (verbose)
                    Console.WriteLine("smaller dst_gap, considered to be equal ");
                return 0;
            }

            if (dst_gap > src_gap + 0.000005 * phase_num)
                return 1;
            else
                return 0;
        }

        void fast_probing()
        {
            bool double_check = false, first = true;
            int tmp_num;
            double interval, pre_gap = 0;
            double saved_ptr = 0, saved_abw = 0;

            probing_start_time = get_time();

            delay_num = 0;
            n_phase_probing(phase_num);
            b_bw = get_bottleneck_bw(dst_gap, total_count);

            delay_num = get_delay_num(avg_dst_gap);
            interval = (double)delay_num / 4;
            delay_num /= 2;

            while (true)
            {
                if (gap_comp(tlt_dst_gap, tlt_src_gap) == 0)
                {
                    if (double_check)
                    {
                        tlt_dst_gap = pre_gap;
                        break;
                    }
                    else
                    {
                        pre_gap = tlt_dst_gap;
                        double_check = true;
                    }
                }
                else
                {
                    if (double_check && verbose)
                        Console.WriteLine("DISCARD: gap = {0}", (int)(tlt_dst_gap * 1000000));
                    double_check = false;
                }

                if (!first && tlt_dst_gap >= 1.5 * tlt_src_gap)
                {
                    tmp_num = get_delay_num((tlt_src_gap + tlt_dst_gap) / (2 * total_count));
                    if (tmp_num > delay_num)
                        delay_num = tmp_num;
                }
                delay_num = (int)(delay_num + interval);

                // TODO: need a better way to deal with them
                saved_ptr = PTR_bw;
                saved_abw = a_bw;
                n_phase_probing(phase_num);
                first = false;
            }
            PTR_bw = saved_ptr;
            a_bw = saved_abw;

            probing_end_time = get_time();
            dump_bandwidth();
        }
    }

    public class PTREventArgs : EventArgs
    {
        public double m_PTR_bw
        {
            get { return d_ptr; }

        } private double d_ptr;

        public double m_IGI_bw
        {
            get { return d_igi; }
        } private double d_igi;

        public PTREventArgs(double ptr, double igi)
        {
            d_ptr = ptr;
            d_igi = igi;
        }
    }
}
