/**
 * Copyright (c) 2009 - Lehigh University. Bethlehem, PA, USA.
 * All rights reserved.
 * This source code is a part of MARCHES Project. 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs, and the author attribution appear in all copies of this
 * software.
 *
 * IN NO EVENT SHALL LEHIGH UNIVERSITY BE LIABLE TO ANY PARTY FOR DIRECT, 
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LEHIGH
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 * LEHIGH UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
 * AND LEHIGH UNIVERSITY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 **/

// modsDlg.cpp : implementation file
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Threading;
using System.Timers;
using System.Management;
using Microsoft.Win32;

namespace MassTools.DeviceAwareness.AvailableCPU
{
    public class AvailableCPU : MassTool
    {
        public event MassToolDataReadyEventHandler DataReadyEvent;

        System.Diagnostics.PerformanceCounter _perfCounter = null;

        private System.Timers.Timer testfreqTimer = null;
        public Thread cputestThread = null;
        bool mbStarted = false;
        int m_nInterval = 500;

        Thread m_toolThread; // for test purpose

        public AvailableCPU()
        {
            //"\Process()\% Processor Time" can go up to N*100 (where N is the number of CPUs) because it adds up the CPU usage of the requested process across all the CPUs.
            //"\Process(_Total)\% Processor Time" should always be around N*100 (where N is the number CPUs) because it adds up the CPU usage of each process, including the idle process.
            //"\Processor()\% Processor Time" can go up to 100 because its the CPU usage of the requested CPU.
            //"\Processor(_Total)\% Processor Time" can go up to 100 because its the average CPU usage across all CPUs.

            _perfCounter = new System.Diagnostics.PerformanceCounter("Processor", "% Processor Time", "_Total", true);
            // _perfCounter.CategoryName = "Processor";
            // _perfCounter.CounterName = "% Processor Time";
            // _perfCounter.InstanceName = "_Total";

            testfreqTimer = new System.Timers.Timer();
            testfreqTimer.Elapsed += new ElapsedEventHandler(testfreqTimer_Elapsed);

            cputestThread = new Thread(new ThreadStart(recvfun));

            mbStarted = false;

            m_toolThread = new Thread(new ThreadStart(MeasureFunc)); // for test purpose
        }

        public override bool Start()
        {
            if (mbStarted) return true;

            //testfreqTimer.Interval = m_nInterval;
            //testfreqTimer.Start();

            m_toolThread.Start(); // for test purpose

            mbStarted = true;
            return true;
        }

        public override bool Stop()
        {
            if (!mbStarted) return true;

            //testfreqTimer.Stop();

            m_toolThread.Abort();  // for test purpose

            mbStarted = false;
            return true;
        }

        public void SetInterval(Int32 n)
        {
            m_nInterval = n;
        }

        public void testfreqTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            DataReadyEvent(this, new MassToolEventArgs(_perfCounter.NextValue()));
        }

        public void recvfun()
        {
        }

        public void AddSubscriber(MassToolDataReadyEventHandler handler)
        {
            DataReadyEvent += handler;
        }

        public string GetProcessorID()
        {
            string sCpuInfo = String.Empty;
            bool bSuccess = false;

            //*** Declare Management Class
            ManagementClass clsMgtClass = new ManagementClass("Win32_Processor");
            ManagementObjectCollection colMgtObjCol = clsMgtClass.GetInstances();

            //*** Loop Over Objects
            foreach (ManagementObject objMgtObj in colMgtObjCol)
            {
                //*** Only return cpuInfo from first CPU
                if (sCpuInfo == String.Empty)
                {
                    sCpuInfo = objMgtObj.Properties["ProcessorId"].Value.ToString();
                    bSuccess = true;
                }
            }

            return sCpuInfo;
        }

        public void GetTotalCPUSpeed()
        {
            double totalCapacity = 0;
            ObjectQuery objectQuery = new ObjectQuery("select * from Win32_Processor"); //Win32_PhysicalMemory
            ManagementObjectSearcher searcher = new
            ManagementObjectSearcher(objectQuery);
            ManagementObjectCollection vals = searcher.Get();

            foreach (ManagementObject val in vals)
            {
                totalCapacity += System.Convert.ToDouble(val.GetPropertyValue("CurrentClockSpeed")); //DeviceID, Name
            }

            Console.WriteLine("Current CPU Speed = " + totalCapacity + " MHz");
            Console.WriteLine("Current CPU Speed = " + (totalCapacity / 1024) + " GHz");
        }

        public void GetCPURegkeyInfo()
        {
            RegistryKey RegKey = Registry.LocalMachine;
            RegKey = RegKey.OpenSubKey("HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0");
            Object cpuSpeed = RegKey.GetValue("~MHz");
            Object cpuType = RegKey.GetValue("VendorIdentifier");
            Console.WriteLine("You have a {0} running at {1} MHz.", cpuType, cpuSpeed);
        }

        //for test purpose
        public void MeasureFunc()
        {
            Thread.Sleep(4000);
            DataReadyEvent(this, new MassToolEventArgs(0.7));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(1.3));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(0.62));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(1.78));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(0.73));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(1.75));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(0.25));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(1.20));
            Thread.Sleep(8000);
            DataReadyEvent(this, new MassToolEventArgs(0.65));
        }
    }
}
